__all__ = ['Predictor']


class Predictor:
    """Parent type of the individual predictors on the Citrine Platform.

    The Predictor class unifies the composite GraphPredictor and individual PredictorNode types
    but does not provide any functionality on its own.

    """
